
// Its a component of the library
#define BUILDING_LIBRARY

// Include hardware info for this device	
#include <avr/io.h>

		
	
#include <libdefs.h>	
#include <servos.h>	
		
		
#define TIMER_PRESCALE 			8		 
#define TIMER_PRESCALE_INDEX 	2		 
#define TIMER_MODE 				TIMER_MODE_NORMAL
#define ON_INT_PENDING bit_is_set(TIFR1,OCF1A)		
#define ON_INT_CLEAR sbi(TIFR1,OCF1A)		
		
#define OFF_INT_PENDING bit_is_set(TIFR1,OCF1B)		
#define OFF_INT_CLEAR sbi(TIFR1,OCF1B)		
		
#define ON_INT_ENABLE sbi(TIMSK1,OCIE1A)		
#define ON_INT_DISABLE cbi(TIMSK1,OCIE1A)		
		
#define OFF_INT_ENABLE sbi(TIMSK1,OCIE1B)		
#define OFF_INT_DISABLE cbi(TIMSK1,OCIE1B)		
		
#define ON_THRESHOLD OCR1A
#define OFF_THRESHOLD OCR1B
		
#define NOW TCNT1		


// The driver class for setting the speed etc
extern __ACTUATOR_DRIVER_CLASS c_Software_Servo;

// The current driver
static SERVO_DRIVER* _driver;
static PORT port;
static PIN mask;

// Channel A is used to set the pin for the next servo
ISR(TIMER1_COMPA_vect){
	SERVO_DRIVER* driver = _driver;
	uint16_t delay;	// The next delay in cycles

	// Wait until previous pulse has been turn off
	while( mask ){
		INTERRUPTABLE_SECTION{
			nop();
		}
	}

	// Start processing the next one
	if(! driver->servos_init){
		// Initialise
		driver->servos_init=TRUE;
		driver->specific.softwarePWM.currentServo = driver->num_servos;
		// Set the threshold to when the first compare was fired
		// so that if there are multiple banks then each bank is
		// reset to use its actual start time.
		ON_THRESHOLD = NOW;
	}else{
		// Do the next servo
		driver->specific.softwarePWM.currentServo++;

		if(driver->specific.softwarePWM.currentServo != driver->num_servos){
			// We are still processing
			if(driver->specific.softwarePWM.currentServo > driver->num_servos){
				// restart from the beginning
				driver->specific.softwarePWM.currentServo = 0;
				driver->specific.softwarePWM.total_delay = 0;
			}

			// Locate the servo
			SERVO* servo = (SERVO *)pgm_read_word(&driver->servos[driver->specific.softwarePWM.currentServo]);
			delay = servo->delay;		// The pulse length to set

			uint16_t now = NOW;

			// ignore the servo if its not connected
			if(act_isConnected(servo)){
				// turn on servo channel bit
				const IOPin* io = servo->pin;
				const IOPort* portDef = (const IOPort*)pgm_read_word(&io->port);
				// Save port and mask for turing off
				port = pgm_read_word(&portDef->port);
				mask = pgm_read_byte(&io->pin);
				_SFR_MEM8(port) |= mask;

				// Set time to turn pin low
				OFF_THRESHOLD = servo->delay + now;
				OFF_INT_CLEAR;
				OFF_INT_ENABLE;
			}

			// Set the time to check the next servo
			ON_THRESHOLD = servo->max_ticks + now;

			// Accumulate total time in this pass of the servos
			driver->specific.softwarePWM.total_delay += servo->max_ticks;

		}else{
			// We have processed all servos

			// disable channel that turns pulses off
			OFF_INT_DISABLE;

			// pause before going back to the first servo
			delay = (driver->specific.softwarePWM.total_delay > driver->specific.softwarePWM.servo_cycle)
					? 100
					:  driver->specific.softwarePWM.servo_cycle - driver->specific.softwarePWM.total_delay;

			// Re-fire the event
			ON_THRESHOLD += delay; 
		}
	}
}


// Channel B is used to turn off the current pulse
ISR(TIMER1_COMPB_vect){
	if(mask){
		// turn off the previous channel
		_SFR_MEM8(port) &= ~mask;
		mask = 0;
	}
	// disable any more of these interrupts
	OFF_INT_DISABLE;
}

/**
 * Should already have checked that Timer is 16bit
 * and has more than two compare channels
 * mode = TIMER_MODE_NORMAL
 */

void bank_servo_Init(SERVO_DRIVER* driver){
	// Set the driver
	_driver = driver;

	// Find how many timer ticks there will be for SERVO_CYCLE
	driver->specific.softwarePWM.servo_cycle = (SERVO_CYCLE * F_CPU) / (TIMER_PRESCALE * 1000);

	// Make sure each servo is set as an output
	for(uint8_t i = 0; i < driver->num_servos; i++){
		SERVO* servo = (SERVO*)pgm_read_word(&driver->servos[i]);

		// Connect the servo to this driver
		servo->actuator.sclass = &c_Software_Servo;
		servo->driver = driver;
		servo->top = driver->specific.softwarePWM.servo_cycle;

		// Make it into an output pin
		pin_make_output(servo->pin, FALSE);

		// Set up center and range
		servoSetConfig(servo,servo->center_us,servo->range_us);

		// Re-connect the servo
		act_setConnected(servo,TRUE);

		// Start off braking
		act_setSpeed(servo,DRIVE_SPEED_BRAKE);
	}


	// Set the threshold to start interrupt chain
	if(getError()==0){
		CRITICAL_SECTION{
//			OFF_INT_DISABLE;	// Is disabled by default
			ON_INT_CLEAR;
			ON_INT_ENABLE;
		}
	}
}
	